//****************************************************************************
// (C) 2013 Alexandre Suaide
// Copyright (C) 2000-2005  ARW Elektronik Germany
//
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// This product is not authorized for use as critical component in 
// life support systems without the express written approval of 
// ARW Elektronik Germany.
//  
// Please announce changes and hints to ARW Elektronik
//
// Maintainer(s): Klaus Hitschler (klaus.hitschler@gmx.de)
//
//****************************************************************************

//****************************************************************************
//
// pcicc32.c -- the driver module for the PCICC32 PCI to CAMAC Interface
//              Thanks to A.Rubini's Book and Dirk Muelhlenberg and H.J.Mathes 
//              for their arwvme driver
//
// $Log: main.c,v $
// Revision 1.27  2005/10/08 13:21:52  klaus
// release 6.7, removed "requested io-memory still claimed" bug at incomplete termination
//
// Revision 1.26  2005/10/07 16:57:10  klaus
// fixed a bug with request_irq with IRQs greater than 127
//
// Revision 1.25  2005/03/11 13:23:26  klaus
// simple corrections for to use with kernels 2.4.21
//
// Revision 1.24  2004/08/12 19:59:19  klaus
// conversion to kernel-version 2.6, released version 6.0
//
// Revision 1.23  2004/01/16 18:42:26  klaus
// converted remap_page_range call for kernels >= 2.4.20, contributed by Roberto Bertoni,
// release of version 5.3
//
// Revision 1.22  2003/12/04 20:34:49  klaus
// minor change: restore the hardware set intCSR content at release of path, release of version 5.2
//
// Revision 1.21  2003/06/19 08:23:38  klaus
// re-compiled with RH-7.2 (kernel 2.4.10)
//
// Revision 1.20  2003/05/12 20:59:50  klaus
// another improvement at previous place
//
// Revision 1.19  2003/05/12 20:38:28  klaus
// improved debug messages from test of connection
//
//****************************************************************************

#define VERSION_HI 6
#define VERSION_LO 7

/*--- INCLUDES ---------------------------------------------------------------------------*/
#include "common.h"  /* must be the first include */

//#include <linux/config.h>

#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/pci.h>
#include <asm/types.h>
#include <linux/delay.h>
#include <linux/wait.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/interrupt.h>
#endif

#include "list.h"
#include "askpci.h"
#include "plxbug.h"
#include "plx9050.h"
#include "fops.h"
#include "pcicc32.h"
#include "pelletron.h"

/*--- INCLUDES -----------------------------------------------------------------------------------*/
#include "common.h"  /* must be the first include */
#include "defines.h"  /* must be the first include */

#include <linux/kernel.h> /* printk() */
#include <linux/mm.h>
#include <linux/module.h> /* MODULE_AUTHOR and MAJOR ... */
#include <linux/pci.h>
#include <asm/errno.h>
#include <asm/types.h>
#include <asm/uaccess.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/poll.h>

#include "fops.h"
#include "list.h"
#include "plx9050.h"
#include "pcicc32.h"      /* the common ioctl commands and structures between driver and application */


/*--- DEFINES ------------------------------------------------------------------------------------*/
MODULE_AUTHOR("klaus.hitschler@gmx.de");
MODULE_DESCRIPTION("Driver for ARW Elektronik PCI CAMAC interface.");
MODULE_SUPPORTED_DEVICE("PCICC32");

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,10)
MODULE_LICENSE("GPL");               // since 2.4.?
#endif

#ifndef MINOR
#define MINOR(x) minor(x)            // since 2.5.?
#endif

#define MAKE_CC32_ADR(N, A, F) (u16)((N << 10) + (A << 6) + ((F & 0xf) << 2))

#define ENABLE_PCIADA_IRQS  (u16)0x0049 /* enable PCIADA IRQs            */
#define DISABLE_PCIADA_IRQS (u16)0x0009 /* disable PCIADA IRQs           */


/*--- DEFINES -----------------------------------------------------------------------------*/
#define MAJOR_NO                       26                 
#define DEVICE_NAME                    "pcicc32"

#if (VERSION_HI < 3)
#define ARWVME_DEVICE_ID               0x9050
#define ARWVME_VENDOR_ID               0x10B5
#define ARWVME_SUBSYSTEM_ID            0x1167
#define ARWVME_SUBSYSTEM_VENDOR_ID     0x9050
#else
#define ARWVME_DEVICE_ID               0x2258
#define ARWVME_VENDOR_ID               0x10B5
#define ARWVME_SUBSYSTEM_ID            0x2258
#define ARWVME_SUBSYSTEM_VENDOR_ID     0x9050
#endif

#define MAKE_CC32_ADR(N, A, F) (u16)((N << 10) + (A << 6) + ((F & 0xf) << 2))

/*--- TYPEDEFS ----------------------------------------------------------------------------*/
extern struct file_operations pcicc32_fops;

/*--- DRIVER GLOBALS ----------------------------------------------------------------------*/
static List *pci_device_header          = NULL;  /* list of all PCIADAs found */
       List *pcicc32_work_device_header = NULL;  /* list of working device interfaces */
static unsigned int nMajor;

/*--- FUNCTIONS ---------------------------------------------------------------------------*/
static int my_interrupt(u16 intCSR)
{
  int result = NOT_MY_INTERRUPT;
  
  if (intCSR & 0x0040)  // it is global enabled
  {    
    if ((intCSR & 0x0028) == 0x0028) // it is a enabled PCIADA interrupt
      result = PCIADA_INTERRUPT;
    else
      if ((intCSR & 0x0005) == 0x0005) // it is a enabled CC32 interrupt
        result = CC32_INTERRUPT;   
  }

  return result;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30)
static irqreturn_t cc32_irqhandler(int irq, void *dev_id)
#else
static irqreturn_t cc32_irqhandler(int irq, void *dev_id, struct pt_regs *regs)
#endif
{
  CC32_DESCRIPTOR *wd = (CC32_DESCRIPTOR *)dev_id;
  
  if (wd)
  {
    // evaluate the reason of the interrupt - if it is mine
    u16 intCSR          = readw((const volatile void*)wd->pLCR + PLX9050_INTCSR);
	  int which_interrupt = my_interrupt(intCSR); 
	  
	  if (which_interrupt)
	  {
	    writew(intCSR & ~0x40, (volatile void*)wd->pLCR + PLX9050_INTCSR);       /* disable global interrupts */
	    
	    read_event_now();
	    
	    wd->wIrqStatus = (u16)which_interrupt;
	    wd->dwInterruptCount++;
	    wake_up_interruptible(&wd->event_queue);                 /* stop blocking if any */
	  }  
	  
	  return IRQ_RETVAL(1);
  }
  
  return IRQ_RETVAL(0);
}

static int request_io_memory(PCIConfigHeader *pPch)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
	if (check_mem_region(pPch->desc[0].base_address, pPch->desc[0].size))
		return -EBUSY;
	
	if (check_mem_region(pPch->desc[3].base_address, pPch->desc[3].size))
		return -EBUSY;
	
	request_mem_region(pPch->desc[0].base_address, pPch->desc[0].size, DEVICE_NAME);
	request_mem_region(pPch->desc[3].base_address, pPch->desc[3].size, DEVICE_NAME);
#endif
	
	return 0;
}

static void release_io_memory(PCIConfigHeader *pPch)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,0)
	release_mem_region(pPch->desc[0].base_address, pPch->desc[0].size);
	release_mem_region(pPch->desc[3].base_address, pPch->desc[3].size);
#endif
}

static int translate_addresses(CC32_DESCRIPTOR *pWd, PCIConfigHeader *pPch)    /* differs from PCIVME */
{
	if (pPch->desc[0].type == PCI_BASE_ADDRESS_MEM_TYPE_1M)  /* LCR ISA base addresses */
		pWd->pLCR = (u32)bus_to_virt(pPch->desc[0].base_address);
	else
		pWd->pLCR = (u32)ioremap(pPch->desc[0].base_address, pPch->desc[0].size);

	if (pPch->desc[3].type == PCI_BASE_ADDRESS_MEM_TYPE_1M)  /* User ISA base addresses */
		pWd->pUsr = (u32)bus_to_virt(pPch->desc[3].base_address);
	else
		pWd->pUsr = (u32)ioremap(pPch->desc[3].base_address, pPch->desc[3].size);
	
	return 0;
}

static void un_translate_addresses(CC32_DESCRIPTOR *pWd, PCIConfigHeader *pPch)
{
	if (pPch->desc[0].type != PCI_BASE_ADDRESS_MEM_TYPE_1M)  /* no LCR ISA base addresses */
		iounmap((void *)pWd->pLCR);
	if (pPch->desc[3].type != PCI_BASE_ADDRESS_MEM_TYPE_1M)  
		iounmap((void *)pWd->pUsr);
}

static void soft_init(CC32_DESCRIPTOR *pd)
{
	if (pd)
	{
    init_waitqueue_head(&pd->event_queue);
    
		pd->pLCR = pd->pUsr = 0;
		pd->pPch = (PCIConfigHeader *)NULL;
		pd->cModuleNumber = 255;
		pd->cFPGAVersion = 255;
		pd->bConnected = 0;
		pd->wInitStep = 0;     
	  pd->wIrq = 0xFFFF;          
	  pd->dwInterruptCount = 0;
	  pd->wIrqStatus = 0;  
	}
}

int test_connection(CC32_DESCRIPTOR *wd)
{
	u16 intCSR_store;
	u16 cntrl_store;
	int i = 1000;
	u32 access_store;
	int error = 0;
	__u32 dwData;
	__u32 dwAdr = wd->pUsr + MAKE_CC32_ADR(26,0,0);
	
	cntrl_store  = readw((const volatile void*)wd->pLCR + PLX9050_CNTRL);  /* read CONTROL register */
	intCSR_store = readw((const volatile void*)wd->pLCR + PLX9050_INTCSR); /* read interrupt + CSR register */
	
	writew(0, (volatile void*)wd->pLCR + PLX9050_INTCSR); /* disable interrupts */
	writew(cntrl_store | 0x0180, (volatile void*)wd->pLCR + PLX9050_CNTRL); /* enable access */
	
	access_store = readl((const volatile void*)dwAdr);
	while (i--)
	{
		writel(0x55555555, (volatile void*)dwAdr);
		dwData = readl((const volatile void*)dwAdr) & 0x00FFFFFF;
		if (0x00555555 != dwData)
		{
			DPRINTK(KERN_DEBUG "pcicc32 : write 0x55555555, read 0x%08x\n", dwData);
			error = 1;
			break;
		}
		writel(0xAAAAAAAA, (volatile void*)dwAdr);
		dwData = readl((const volatile void*)dwAdr) & 0x00FFFFFF;
		if (0x00AAAAAA != dwData)
		{
			DPRINTK(KERN_DEBUG "pcicc32 : write 0xAAAAAAAA, read 0x%08x\n", dwData);
			error = 1;
			break;
		}
		writel(0xFFFFFFFF, (volatile void*)dwAdr);
		dwData = readl((const volatile void*)dwAdr) & 0x00FFFFFF;
		if (0x00FFFFFF != dwData)
		{
			DPRINTK(KERN_DEBUG "pcicc32 : write 0xFFFFFFFF, read 0x%08x\n", dwData);
			error = 1;
			break;
		}
		writel(0x00000000, (volatile void*)dwAdr);
		dwData = readl((const volatile void*)dwAdr) & 0x00FFFFFF;
		if (0x00000000 != dwData)
		{
			DPRINTK(KERN_DEBUG "pcicc32 : write 0x00000000, read 0x%08x\n", dwData);
			error = 1;
			break;
		}
	}
	
	writew(cntrl_store & ~0x0100, (volatile void*)wd->pLCR + PLX9050_CNTRL); /* clear potential interrupt */
	
	/* restore all contents */
	writel(access_store, (volatile void*)dwAdr);
	writew(cntrl_store, (volatile void*)wd->pLCR + PLX9050_CNTRL);
	writew(intCSR_store, (volatile void*)wd->pLCR + PLX9050_INTCSR);
	
	return error; 	
}

int get_module_info(CC32_DESCRIPTOR *wd, u8 *cModuleNumber, u8 *cFPGAVersion)
{
	u16 intCSR_store;
	u16 cntrl_store;
	int found = 0;
	u16 data;
	
	cntrl_store  = readw((const volatile void*)wd->pLCR + PLX9050_CNTRL);  /* read CONTROL register */
	intCSR_store = readw((const volatile void*)wd->pLCR + PLX9050_INTCSR); /* read interrupt + CSR register */
	
	DPRINTK(KERN_DEBUG "pcicc32 : get_module_info(), cntrl=0x%04x, intCSR=0x%04x\n", cntrl_store, intCSR_store); 
	
	if (cntrl_store & 0x0800) /* a CC32 is connected */
	{
		u16 bla = cntrl_store | 0x0180;
		
		writew(0, (volatile void*)wd->pLCR + PLX9050_INTCSR); /* disable interrupts */
		writew(bla, (volatile void*)wd->pLCR + PLX9050_CNTRL); /* enable access */
		
		data = readw((const volatile void*)(wd->pUsr + MAKE_CC32_ADR(0,0,0)));
		  
		DPRINTK(KERN_DEBUG "pcicc32 : CC32 status=0x%04x\n", data); 
	
		if ((data & 0xF000) != 0x8000)
		{
			*cModuleNumber = *cFPGAVersion = 255;
			printk(KERN_ERR "pcicc32 : Wrong module type connected @ index %d (0x%04x)!\n", wd->wIndex, data);
		}
		else
		{
			found = 1;
			*cModuleNumber = (data >> 4) & 0xF;
			*cFPGAVersion  = (data >> 8) & 0xF; 
		} 
	
		writew(cntrl_store & ~0x0080, (volatile void*)wd->pLCR + PLX9050_CNTRL); /* clear potential interrupt */
	
		/* restore all contents */
		writew(cntrl_store, (volatile void*)wd->pLCR + PLX9050_CNTRL);
		writew(intCSR_store, (volatile void*)wd->pLCR + PLX9050_INTCSR);
	}
	
	return found; 	
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,19)
static int pcicc32_read_proc(char *buf, char **start, off_t offset, int len)
#else
static int pcicc32_read_proc(char *buf, char **start, off_t offset, int len, int *eof, void *data)
#endif
{
  int i;
  int pos = 0;
  CC32_DESCRIPTOR *pd;
  PCIConfigHeader *ch;
  Node            *n;
  u16 		        cntrl;
  char            *cause = "None";
  
	DPRINTK(KERN_DEBUG "pcicc32 : pcicc32_read_proc()\n"); 
	
  pos += sprintf(buf + pos, "\npcicc32 information. Version %d.%d of %s from Klaus Hitschler.\n", VERSION_HI, VERSION_LO, __DATE__);
  
  i = getNumOfNodesInList(pcicc32_work_device_header);
  pos += sprintf(buf + pos, " ---------------------\n"); 
  pos += sprintf(buf + pos, " Interfaces found : %d\n", i);
  
  n = getFirstNode(pcicc32_work_device_header);
  while (i--)
  {
  	pd = (CC32_DESCRIPTOR *)getContent(n);
  	ch = pd->pPch;
	  cntrl = readw((const volatile void*)pd->pLCR + PLX9050_CNTRL);
          pos += sprintf(buf + pos, " --- %d ---------------\n", i + 1); 
	  pos += sprintf(buf + pos, " LCR  phys/virt/size : 0x%08x/0x%08x/%d\n", ch->desc[0].base_address, pd->pLCR, ch->desc[0].size);
	  pos += sprintf(buf + pos, " User phys/virt/size : 0x%08x/0x%08x/%d\n", ch->desc[3].base_address, pd->pUsr, ch->desc[3].size);
	  pos += sprintf(buf + pos, " Irq                 : %d\n", pd->wIrq);
	  
	  if (pd->bConnected)
	  {
		  pos += sprintf(buf + pos, " CC32 is or was      : (software) connected.\n");
	  	pos += sprintf(buf + pos, " Module-Number       : %d\n", pd->cModuleNumber);
	  	pos += sprintf(buf + pos, " FPGA-Version        : %d\n", pd->cFPGAVersion);
		}
		else
		  pos += sprintf(buf + pos, " CC32 is or was      : not (software) connected.\n");		
		
	  if (!((cntrl & 0x0800) && (!(cntrl & 0x0600))))
	  	pos += sprintf(buf + pos, " CC32 is             : powered off or cable disconnected.\n");
		
	  pos += sprintf(buf + pos, " IrqCount            : %d\n", pd->dwInterruptCount);
	  if (pd->wIrqStatus & PCIADA_INTERRUPT)
	    cause = "Timeout";
	  else
	    if (pd->wIrqStatus & CC32_INTERRUPT)
	      cause = "LAM";	  
	  pos += sprintf(buf + pos, " Pending IrqStatus   : %s\n", cause);
	  
  	n = getNextNode(n);
  }

  pos += sprintf(buf + pos, "\n");
  
  #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,19)
  *eof = 1;
  #endif
  
  return pos;
}

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,2,19)
//----------------------------------------------------------------------------
// replacement for kernel 2.4.x function
// original from sysdep.h out from the book
// "Linux Device Drivers" by Alessandro Rubini and Jonathan Corbet, published by O'Reilly & Associates.
static struct proc_dir_entry *create_proc_read_entry(const char *name, mode_t mode, 
                                        struct proc_dir_entry *base, read_proc_t *read_proc, void * data)
{
  struct proc_dir_entry *res = create_proc_entry(name, mode, base);
  if (res) 
  {
    res->read_proc=read_proc;
    res->data=data;
  }
  return res;
}
#endif


int init_module(void)
{
  int i = 0;
  PCIConfigHeader *ch;
  CC32_DESCRIPTOR *wd = NULL;
  Node            *n;
  int             result = 0;
  
  printk(KERN_INFO "pcicc32 : init_module\n");

  /* create list of PCIADAs and work devices */
  pci_device_header          = newList();
  pcicc32_work_device_header = newList();
  
  /* search for all PCIADA modules */
  while ((ch = GetPCIConfigHeader(ARWVME_VENDOR_ID, ARWVME_DEVICE_ID, i)))
  {
    if (ch->subsystem_id != ARWVME_SUBSYSTEM_ID) 
    {
      printk(KERN_ERR "pcicc32 : found 0x%x, 0x%x , but wrong subsystem id 0x%04x!\n",
              ARWVME_VENDOR_ID, ARWVME_DEVICE_ID, ch->subsystem_id);
      i++;
      kfree_s(ch, sizeof(*ch));        // FREE(ch);
      continue;
    }

    if (ch->subsystem_vendor_id != ARWVME_SUBSYSTEM_VENDOR_ID) 
    {
      printk(KERN_ERR "pcicc32 : found 0x%x, 0x%x , but wrong subsystem vendor Id 0x%04x!\n",
             ARWVME_VENDOR_ID, ARWVME_DEVICE_ID, ch->subsystem_vendor_id);
      i++;
      kfree_s(ch, sizeof(*ch));        // FREE(ch);
      continue;
    }

    /* test if it is configured as PCICC32 or a PCIVME interface card */
    if (ch->desc[3].size != 0x8000)  
    {
      printk(KERN_ERR "pcicc32 : found, but wrong memory window size 0x%08x!\n", ch->desc[3].size);
      i++;
      kfree_s(ch, sizeof(*ch));        // FREE(ch);
      continue;
    }

    printk(KERN_INFO "pcicc32 : found %d. PCIADA card\n", i + 1);
    addTail(pci_device_header, ch);  /* add this header to list */
    i++;
  }
  
  /* fix the PLX bug in all PCIADAs */
  i = getNumOfNodesInList(pci_device_header);
  n = getFirstNode(pci_device_header);
  while (i--)
  {
  	ch = (PCIConfigHeader *)getContent(n); 
  	PLX9050BugFix(ch);
  	n = getNextNode(n);
  }

  /* create work_devices and translate the access addresses */
  i = getNumOfNodesInList(pci_device_header);
  n = getFirstNode(pci_device_header);
  while (i--)
  {
  	wd = (CC32_DESCRIPTOR *)kmalloc(sizeof(CC32_DESCRIPTOR), GFP_ATOMIC);
	  soft_init(wd);
	  ch  = (PCIConfigHeader *)getContent(n);
	  wd->pPch = ch;
	  wd->wIndex = i;
	
	  if (!request_io_memory(ch))
	  {
	    // successful request_io_memory
	    wd->wInitStep = 1;
	    
		  if (translate_addresses(wd, ch))
		  {
			  printk(KERN_ERR "pcicc32 : translation of addresses failed!\n");
        release_io_memory(ch);
			  kfree_s(wd, sizeof(*wd));       // FREE(wd);
		  }
		  else
		  {
		    // successful translate_addresses
		    wd->wInitStep = 2;
		    
        if (request_irq(wd->pPch->PCI_dev->irq, cc32_irqhandler, IRQF_DISABLED | IRQF_SHARED, DEVICE_NAME, wd))
	      {
				  printk(KERN_ERR "pcicc32 : can't get irq @ %d\n", wd->pPch->PCI_dev->irq);
          un_translate_addresses(wd, ch);
          release_io_memory(ch);
			    kfree_s(wd, sizeof(*wd));       // FREE(wd);
	      }
	      else
	      {   
	        // successful request_irq
		      wd->wInitStep = 3;
		      wd->wIrq = wd->pPch->PCI_dev->irq;
		
			    addTail(pcicc32_work_device_header, wd);  /* add this device to list */
			    wd->bConnected =  get_module_info(wd, &wd->cModuleNumber, &wd->cFPGAVersion);
			    if (wd->bConnected && test_connection(wd))
			    {
				    printk(KERN_ERR "pcicc32 : connection test failed!\n");
				    wd->bConnected = 0;
			    }
			  }
		  }
	  }
	  else
	  {
		  printk(KERN_ERR "pcicc32 : requested io-memory still claimed!\n");
			kfree_s(wd, sizeof(*wd));       // FREE(wd);
		}
	
  	n = getNextNode(n);
  }
  
  nMajor = MAJOR_NO;
  result = register_chrdev(nMajor, DEVICE_NAME, &pcicc32_fops);
  if (result < 0)
  {
  	printk(KERN_ERR "pcicc32: Can't install driver (%d)\n", result);
	
    /* untranslate translated addresses */
    i = getNumOfNodesInList(pcicc32_work_device_header);
    n = getFirstNode(pcicc32_work_device_header);
    while (i--)
    {
	    wd  = (CC32_DESCRIPTOR *)getContent(n);
	    ch = wd->pPch;
	    un_translate_addresses(wd, ch); 
	    n = getNextNode(n);
    }

    /* delete my lists */
    deleteList(pcicc32_work_device_header, (void (*)(void *))NULL);  
    deleteList(pci_device_header, (void (*)(void *))NULL);   	
  
  	return result;
  }
  else
  {
  	if (nMajor == 0) nMajor = result;
    printk(KERN_INFO "pcicc32 : major #%d assigned.\n", nMajor);
  }
    
  /* register the proc device */ 
  
  CURRENT = wd;
  return create_proc_read_entry(DEVICE_NAME, 0, NULL, pcicc32_read_proc, NULL) ? 0 : -ENODEV; 
}

void cleanup_module(void)
{
  int i;
  PCIConfigHeader *ch;
  CC32_DESCRIPTOR *wd;
  Node            *n;
  
  unregister_chrdev(nMajor, DEVICE_NAME);
  
  DPRINTK(KERN_DEBUG "pcicc32 : cleanup_module()\n");
		
  /* unregister the proc device */
  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,2,19)
  proc_unregister(&proc_root, pcimod_proc_entry.low_ino);
  #else
  remove_proc_entry(DEVICE_NAME, NULL);
  #endif
    
  /* untranslate translated addresses */
  i = getNumOfNodesInList(pcicc32_work_device_header);
  n = getFirstNode(pcicc32_work_device_header);
  while (i--)
  {
	  wd  = (CC32_DESCRIPTOR *)getContent(n);
  	ch = wd->pPch;
	  switch (wd->wInitStep)
	  {
	    case 3:  writew(readw((const volatile void*)wd->pLCR + PLX9050_INTCSR) & ~0x40,
			    (volatile void*)wd->pLCR + PLX9050_INTCSR);  // disable global interrupts
               free_irq(wd->wIrq, wd);  
	    case 2:  un_translate_addresses(wd, ch);
	    case 1:  release_io_memory(ch);
	    default: wd->wInitStep = 0; 
	  }
  	n = getNextNode(n);
  }
  
  /* delete my lists */
  deleteList(pcicc32_work_device_header, (void (*)(void *))NULL);  
  deleteList(pci_device_header, (void (*)(void *))NULL); 
  
  printk(KERN_INFO "pcicc32 : cleanup_module successful.\n");
  
  return;
}

//######################################################################################
//######################################################################################
//######################################################################################
//######################################################################################

void read_event_now()
{
  static unsigned short ev_buf[MAX_EVENT_SIZE];
  int i = 0;
  iev=0;
  lock=0;
  
  ev_read(ev_buf);

  if(iev==0) return;
  
  if(iev==1 && ev_buf[iev-1]==0xFFFF)
  {
    iev = 0;
    return;
  }
  
  if(ev_buf[iev-1]!=0xFFFF) 
  {
    printk("WARNING: No end of event 0xFFFF detected. Fixing it!\n");
    ev_buf[iev] = 0xFFFF;
    iev++;
  }
  
  if(iev>EVENT_SIZE) 
  {
    printk("WARNING: This event has size equal to EVENT_SIZE = %d iev = %d NBUF = %d NEXTW = %d\n",EVENT_SIZE,iev,NBUF,NEXTW);
    iev=EVENT_SIZE;
    ev_buf[iev-1]=0xFFFF;
  }

  if(nwords+iev+4>=MAXSIZE)
  {
    FLAG[NEXTW++]=nwords;
    for(i=nwords;i<MAXSIZE;i++) cc_buf[offset_b+i-nwords] = 0;
    if(NEXTW>=NBUF) NEXTW=0;
    nwords=0;
    offset_b=NEXTW*MAXSIZE;
    if(NBUFFULL<NBUF) NBUFFULL++;
  }
  if(FLAG[NEXTW]==0 && nwords+iev+4<MAXSIZE)
  {
    nwords=nwords+iev;  
    memcpy((void*)&cc_buf[offset_b],(void*)&ev_buf[0],iev*sizeof(unsigned short int));
    offset_b=offset_b+iev;
  }
  else lost++;
  return;
}
/***************************************************************/
int ev_read(unsigned short int ev_buf[])
{  
  int i,j,looping;
  unsigned short int par,PAT=0,EX=0,TXR=0;
  unsigned int inst;
  int hbit;
  unsigned int CA=0;
 
  looping=1;
  i=0;
  j=0;
  while ((looping == 1) && (i<INSTR_SIZE) )
  {
    hbit=(cc_inst_buf[i] & 0x00f00000);
    if(hbit>0x200000)
    {
      inst = (cc_inst_buf[i] & 0x007f0000);
      par  = cc_inst_buf[i++] & 0xffff;
    }
    else
    {
      inst = (cc_inst_buf[i] & 0x007fc000);
      par  = cc_inst_buf[i++] & 0x3fff;
    }
    if(inst == 0x5C0000) par = par|0x0fff;
    if((inst == 0x100000))
    {
        #include "NAF_PCICC32_EVO.code"           
     }
    else if(inst == 0x580000)
    {
      if(iev<EVENT_SIZE) ev_buf[iev++] = par; 
      else ev_buf[EVENT_SIZE-1] = 0xFFFF; 
    }          /* OUT # */
    else if(inst == 0x590000)
    {
      if(iev<EVENT_SIZE) ev_buf[iev++] = (CA&0x0fff); 
      else ev_buf[EVENT_SIZE-1] = 0xFFFF;
      
    }  /* OUT CA */
    else if(inst == 0x400000)CA = par;                                          /* MOV #,CA */
    else if(inst == 0x410000)PAT = CA;                                          /* MOV CA,PAT */
    else if(inst == 0x420000)CA = PAT;                                          /* MOV PAT,CA */
    else if(inst == 0x200000)looping=0;                                         /* RTRN */
    else if(inst == 0x250000)i = par-1;                                         /* BRU A */
    else if(inst == 0x2A0000)looping=0;                                         /* INIT */
    else if(inst == 0x300000){if((PAT&par) != 0)i++;}                           /* SKIP PAT .ANY.# */
    else if(inst == 0x320000){if((EX&par) != 0)i++;}                            /* SKIP EX .ANY.# */
    else if(inst == 0x330000){if((CA&par) != 0)i++;}                            /* SKIP CA .ANY.# */
    else if(inst == 0x340000){if((PAT&par) == 0)i++;}                           /* SKIP PAT .NONE.# */
    else if(inst == 0x360000){if((EX&par) == 0)i++;}                            /* SKIP EX .NONE.# */
    else if(inst == 0x370000){if((CA&par) == 0)i++;}                            /* SKIP CA .NONE.# */
    else if(inst == 0x3b0000){if(CA<par) i++;}                                  /* SKIP CA .LT.# */
    else if(inst == 0x3f0000){if(CA>par) i++;}                                  /* SKIP CA .GT.# */
    else if(inst == 0x380000){if(PAT<par) i++;}                                 /* SKIP PAT .LT.# */
    else if(inst == 0x3c0000){if(PAT>par) i++;}                                 /* SKIP PAT .GT.# */
    else if(inst == 0x430000)EX = par;                                          /* SETX # */
    else if(inst == 0x440000)EX = EX + par;                                     /* ADDX # */
    else if(inst == 0x450000)CA = CA >> par;                                    /* SHFT # */
    else if(inst == 0x510000)TXR = CA;                                          /* MOV CA,TXR */
    else if(inst == 0x5A0000){if(iev<EVENT_SIZE) ev_buf[iev++] = PAT; else ev_buf[EVENT_SIZE-1] = 0xFFFF;}          /* OUT PAT */
    else if(inst == 0x5C0000){if(iev<EVENT_SIZE) ev_buf[iev++]=(TXR&par); else ev_buf[EVENT_SIZE-1] = 0xFFFF;}      /* MERG #,TXR */
    else if(inst == 0x600000) udelay(par&0x00ff);                               /* DLAY # */
    else if(inst == 0x700000) par = 0;                                          /* CLRB */
    else if(inst == 0x710000) par = 0;                                          /* SETB */
    else if(inst == 0x000000) return iev;                                       /* RETURN */
    else 
    {
      printk(" Invalid instruction \n");
      printk("inst, par: %x %x %x \n",inst,par,cc_inst_buf[i-1]);
    }
  }
  return iev;
}
/*--- EXTERNALS ----------------------------------------------------------------------------------*/
extern int get_module_info(CC32_DESCRIPTOR *wd, u8 *cModuleNumber, u8 *cFPGAVersion);
extern int test_connection(CC32_DESCRIPTOR *wd);
extern List *pcicc32_work_device_header;

/*--- TYPEDEFS -----------------------------------------------------------------------------------*/
typedef struct
{
	CC32_DESCRIPTOR *desc;  /* pointer to my PCIADA & CC32 */
} PRIVATE_DATA;

/*--- FUNCTIONS ----------------------------------------------------------------------------------*/
static void switch_CC32_on(CC32_DESCRIPTOR *desc)
{
	u16 cntrl;
	
	DPRINTK(KERN_DEBUG "pcicc32 : switch_CC32_on()\n");
	
	cntrl  = readw((const volatile void*)desc->pLCR + PLX9050_CNTRL);  /* read CONTROL register */
	cntrl |= 0x0180;
	writew(cntrl, (volatile void*)desc->pLCR + PLX9050_CNTRL); /* enable access */
}

static void switch_CC32_off(CC32_DESCRIPTOR *desc)
{
	u16 cntrl  = readw((const volatile void*)desc->pLCR + PLX9050_CNTRL);  /* read CONTROL register */
	
	DPRINTK(KERN_DEBUG "pcicc32 : switch_CC32_off()\n");
	
	writew(DISABLE_PCIADA_IRQS, (volatile void*)desc->pLCR + PLX9050_INTCSR);  /* disable global interrupts */
	cntrl &= ~0x0100;
	writew(cntrl, (volatile void*)desc->pLCR + PLX9050_CNTRL);       /* disable access */
}

int pcicc32_ioctl(struct inode *inode, struct file *pFile, unsigned int cmd, unsigned long arg)
{
  int crate = MINOR(inode->i_rdev);
  int i,arg1;
  static long idum;
  unsigned short int dummy[MAX_EVENT_SIZE];
  arg1=arg;	
  if (crate>=MAX_CRATES) return -ENODEV;
  switch (cmd) 
  {
    case CCMAXSIZE:
      MAXSIZE=arg;
      if(MAXSIZE<MIN_BUFFER_SIZE)  MAXSIZE=MIN_BUFFER_SIZE;
      if(MAXSIZE>MAX_BUFFER_SIZE)  MAXSIZE=MAX_BUFFER_SIZE;
      EVENT_SIZE = MAXSIZE-2;
      if(EVENT_SIZE>MAX_EVENT_SIZE) EVENT_SIZE=MAX_EVENT_SIZE;
      NBUF=BUFFER_SIZE/MAXSIZE;
      if(NBUF>MAX_BUFFER) NBUF=MAX_BUFFER;
      printk(" maxsize,nbuf = %d %d\n",MAXSIZE,NBUF);
      for(i=0;i<BUFFER_SIZE_ARRAY;i++) cc_buf[i] = 0;
      return MAXSIZE;
      break;
    case CCRESET:
      if(arg == ENABLE || arg==0)
      {					
        printk(" Running...\n");
        IRQCOUNT = 0;
        for(i=0;i<BUFFER_SIZE_ARRAY;i++) cc_buf[i] = 0;
        for(i=0;i<MAX_BUFFER;i++) FLAG[i]=0;
        lost=0;
        nwords=0;
        offset_b=0;
        NEXTW=0;
        NEXTR=0;
        NBUFFULL = 0;
        Running = 1;
        arg1=ENABLE;
 	idum  = -534234;
      }
      if(arg==DISABLE || arg==4) 
      {
        printk(" Stopped...lost %d events\n",lost);
        arg1=DISABLE;
       }
       //if(IRQCRATE)
       //{
       //   if(arg==DISABLE) camcl(IRQCRATE,4);
       //   else if(arg==ENABLE) {CURRENT = IRQCRATE; camcl(IRQCRATE,3);} 
       //}
      return 0;
      break;
    case CCWAITLAM:
      printk("Invalid call: CCWAITLAM \n");
      return -1;
      break;
    case CCGETLOST:
      return lost;
      break;
    case CCGETUSED:
      return NBUFFULL;
      break;
    case CCGETBUF:
      return NBUF;
      break;
    case CCSETCRATE:
      //if(arg<MAX_CRATES) set_crate_number(arg);
      return 0;
    default:			
      NN=arg&0x001f;
      FF=(arg&0x03e0)>>5;
      AA=(arg&0x3c00)>>10;			
      CC_IO_OP=cmd;
      return 0;
  }

}

int pcicc32_mmap(struct file *pFile, struct vm_area_struct *vma)
{
	PRIVATE_DATA *pd = (PRIVATE_DATA *)pFile->private_data;
	
	DPRINTK(KERN_DEBUG "pcicc32 : pcicc32_mmap()\n");
	
	if (pd->desc->pPch->desc[3].type == PCI_BASE_ADDRESS_MEM_TYPE_1M)
	{
		DPRINTK(KERN_DEBUG "pcicc32 : mmap(address=0x%08x size=%d)\n", 
			(u32)virt_to_phys((void *)pd->desc->pUsr), pd->desc->pPch->desc[3].size);

		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,30)
		if (remap_pfn_range(vma, vma->vm_start, 
			(u32)virt_to_phys((void *)pd->desc->pUsr), pd->desc->pPch->desc[3].size, PAGE_SHARED))
			return -EAGAIN;
		#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)  
		if (remap_page_range(vma, vma->vm_start, 
			(u32)virt_to_phys((void *)pd->desc->pUsr), pd->desc->pPch->desc[3].size, PAGE_SHARED))
			return -EAGAIN;
		#else
		if (remap_page_range(vma->vm_start, 
			(u32)virt_to_phys((void *)pd->desc->pUsr), pd->desc->pPch->desc[3].size, PAGE_SHARED))
			return -EAGAIN;
		#endif
	}
	else
	{
		DPRINTK(KERN_DEBUG "pcicc32 : mmap(address=0x%08x size=%d)\n", 
			(u32)virt_to_phys((void *)pd->desc->pUsr), pd->desc->pPch->desc[3].size);

		#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,30)
		if (remap_pfn_range(vma, vma->vm_start, 
			pd->desc->pPch->desc[3].base_address, pd->desc->pPch->desc[3].size, PAGE_SHARED))
			return -EAGAIN;
		#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22)  
		if (remap_page_range(vma, vma->vm_start, 
			pd->desc->pPch->desc[3].base_address, pd->desc->pPch->desc[3].size, PAGE_SHARED))
			return -EAGAIN;
		#else
		if (remap_page_range(vma->vm_start, 
			pd->desc->pPch->desc[3].base_address, pd->desc->pPch->desc[3].size, PAGE_SHARED))
			return -EAGAIN;
		#endif
	}
	
	return 0;
}

int pcicc32_open(struct inode *pInode, struct file *pFile)
{
	CC32_DESCRIPTOR *wd = 0;
	CC32_DESCRIPTOR *desc = 0;
	int i;
	Node *n;
	int nMinor = MINOR(pInode->i_rdev);
	PRIVATE_DATA *pd;
	
	DPRINTK(KERN_DEBUG "pcicc32 : pcicc32_open(), Major=%d, Minor=%d\n", MAJOR(pInode->i_rdev), nMinor);
	
	/* search for device */
	i = getNumOfNodesInList(pcicc32_work_device_header);
	n = getFirstNode(pcicc32_work_device_header);
	
	DPRINTK(KERN_DEBUG "pcicc32 : scanning %d devices\n", i);
	while (i--)
	{
		wd  = (CC32_DESCRIPTOR *)getContent(n);
		
		wd->bConnected =  get_module_info(wd, &wd->cModuleNumber, &wd->cFPGAVersion);
		if (wd->bConnected)
		{
			if  (test_connection(wd))
			{
				printk(KERN_ERR "pcicc32 : connection test for module %d failed!\n", wd->cModuleNumber);
				wd->bConnected = 0;
			}
			else
				if (wd->cModuleNumber == nMinor)
				{
					desc = wd;
					break;
				}
		}
		else
		{
			DPRINTK(KERN_DEBUG "pcicc32 : module %d not connected!\n", nMinor);
		}
			  
		n = getNextNode(n);
	}
	
	if (desc)
	{
		pFile->private_data = (void *)kmalloc(sizeof(PRIVATE_DATA), GFP_ATOMIC);
		if (!pFile->private_data)
			return -ENOMEM;
		
		pd = (PRIVATE_DATA *)pFile->private_data;
		pd->desc = wd;
		
		DPRINTK(KERN_DEBUG "pcicc32 : found CC32 module with number %d.\n", nMinor);
	}
	else
	{
		DPRINTK(KERN_DEBUG "pcicc32 : No CC32 module found.\n");

		return -ENODEV;
	}		
		
	switch_CC32_on(wd);
			
	__MOD_INC_USE_COUNT__;
	return 0;
}

int pcicc32_release(struct inode *pInode, struct file *pFile)
{
	PRIVATE_DATA *pd;

	DPRINTK(KERN_DEBUG "pcicc32 : pcicc32_release()\n");
				
	if (pFile->private_data) 
	{
		pd = (PRIVATE_DATA *)pFile->private_data;
		if (pd && pd->desc)	
		{
			switch_CC32_off(pd->desc);
			pd->desc = 0;
		}
			
		kfree_s(pd, sizeof(*pd));    // FREE(pFile->private_data);
	}
	
	__MOD_DEC_USE_COUNT__;
	return 0;
}

/***************************************************************/
static ssize_t cc_read (struct file *file, char *buffer, size_t size, loff_t *loff)
{
  int result;
  int max,nbytes,QX;
  int Dat;
  unsigned short int *offset;
  //unsigned short int dummy[MAX_EVENT_SIZE];
 
  /*  Check that we can really write to the buffer   */
  max=size>>1;
  if(max>iev) max=iev;
  if(max>BUFFER_SIZE) return -1;
  if((result = access_ok(VERIFY_WRITE,buffer,size))&&(buffer != NULL)) return result;

  switch (CC_IO_OP)
  {
    case CCREAD16:
      QX = cami16(NN,AA,FF,&Dat);		  
      memcpy((void*)buffer, &Dat, size);
      return QX;
      break;
    case CCREAD24:
      QX = cami24(NN,AA,FF,&Dat);
      memcpy((void*)buffer, &Dat, size);
      return QX;
      break;
    case CCREADEVNT:
      if(FLAG[NEXTR] > 0)
      {
        offset = (short int *)cc_buf;
        nbytes = FLAG[NEXTR]*2;
        if(nbytes < 1) lock++;
        FLAG[NEXTR]=0;
        if(NBUFFULL>0) NBUFFULL--;
        offset=offset+MAXSIZE*NEXTR;
	      NEXTR++;
        if(NEXTR >= NBUF) NEXTR = 0;
        memcpy((void*)buffer,(void*)offset,nbytes);
        if(lock > 20) 
        {
	        lock=0;
        }
        return nbytes;
      }
      return 0;
      break;
    default:
      return -1;
  }
}
static ssize_t cc_write (struct file *file, const char *buffer, size_t size, loff_t *loff)
{
  int result, Dat=0;
  unsigned short int dummy[MAX_EVENT_SIZE];
  
  if ((result = access_ok(VERIFY_READ,buffer,size))) return result;

  switch (CC_IO_OP)
  {
    case CCWRITE16:
      copy_from_user(&Dat,buffer,2);
      return camo16(NN,AA,FF,Dat);
      break;
    case CCWRITE24:
      copy_from_user(&Dat,buffer,2);
      return camo24(NN,AA,FF,Dat);
      break;
    case CCLOADEH:
      printk("loading EV-STANDARD\n size = %d",(int)size);
      lockEVO = 0;
      copy_from_user((int *)cc_inst_buf,buffer,size);
      printk("loading %d instructions \n",(int)size/4);
      iev=0;
      ev_read(dummy);
      lockEVO = 1;
      return 0;
      break;
    case CCLOADEHPLUS:
      return 0;
      break;
    default:
      printk("I dont know what to do with command %d\n",CC_IO_OP);
      return -1;
  }
}

/***************************************************************/
int camo16(int N, int A, int F,int D)
{
    int NAF = MAKE_CC32_ADR(N,A,F);
    // writew(D,CURRENT->pUsr+NAF);
    return 0;	
}
/***************************************************************/
int camo24(int N, int A, int F,int D)
{
    int NAF = MAKE_CC32_ADR(N,A,F);
    // writew(D,CURRENT->pUsr+NAF);
    return 0;
}
/***************************************************************/
int cami16(int N, int A, int F, int *D)
{
    int NAF = MAKE_CC32_ADR(N,A,F);
    /* *D = readl(CURRENT->pUsr+NAF); */
    return 0;
}
/***************************************************************/
int cami24(int N, int A, int F, int *D)
{
    int NAF = MAKE_CC32_ADR(N,A,F);
    /* *D = readl(CURRENT->pUsr+NAF); */
    return 0;
}

//----------------------------------------------------------------------------
// is called at poll or select
static unsigned int pcicc32_poll(struct file *pFile, poll_table *wait)
{
	PRIVATE_DATA    *pd = (PRIVATE_DATA *)pFile->private_data;
  CC32_DESCRIPTOR *wd = pd->desc;
  unsigned int mask   = 0;
  
	DPRINTK(KERN_DEBUG "pcicc32 : pcicc32_poll()\n");
	
  poll_wait(pFile, &wd->event_queue, wait);
  
  // return on ops that could be performed without blocking
  if (wd->wIrqStatus)      
    mask |= (POLLIN  | POLLRDNORM);
  
  return mask; 
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
struct file_operations pcicc32_fops =
{
	NULL, 			/* lseek */
	cc_read,			/* read  */
	cc_write,			/* write */
	NULL,			/* readdir */
	pcicc32_poll,   /* poll, since 2.2., replaces select */ 
	pcicc32_ioctl,	/* ioctl */
	pcicc32_mmap,	/* mmap */
	pcicc32_open,	/* open */
	NULL,           /* flush */
	pcicc32_release,/* release */
};
#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30)
struct file_operations pcicc32_fops =
{
	unlocked_ioctl: pcicc32_ioctl,		/* ioctl */
	compat_ioctl:   pcicc32_ioctl,		/* ioctl */
	mmap:           pcicc32_mmap,		/* mmap */
	open:           pcicc32_open,		/* open */
	release:        pcicc32_release,	/* release */ 
	poll:           pcicc32_poll,		/* poll */
	read:           cc_read,
	write:          cc_write,
};
#else
struct file_operations pcicc32_fops =
{
	ioctl:   pcicc32_ioctl,		/* ioctl */
	mmap:    pcicc32_mmap,		/* mmap */
	open:    pcicc32_open,		/* open */
	release: pcicc32_release,	/* release */ 
	poll:    pcicc32_poll,		/* poll */
	read:           cc_read,
	write:          cc_write,
};
#endif
